import click

from vantage6.common import info
from vantage6.common.globals import InstanceType

from vantage6.cli.common.stop import execute_stop, helm_uninstall, stop_port_forward
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS, InfraComponentName


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    default=DEFAULT_SERVER_SYSTEM_FOLDERS,
    help="Search for configuration in system folders instead of user folders. "
    "This is the default.",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    help="Search for configuration in the user folders instead of system folders.",
)
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click.option("--all", "all_stores", flag_value=True, help="Stop all algorithm stores")
def cli_algo_store_stop(
    name: str,
    context: str,
    namespace: str,
    system_folders: bool,
    all_stores: bool,
    sandbox: bool,
):
    """
    Stop one or all running algorithm store(s).
    """
    execute_stop(
        stop_function=_stop_store,
        instance_type=InstanceType.ALGORITHM_STORE,
        infra_component=InfraComponentName.ALGORITHM_STORE,
        stop_all=all_stores,
        to_stop=name,
        namespace=namespace,
        context=context,
        system_folders=system_folders,
        is_sandbox=sandbox,
    )


def _stop_store(store_name: str, namespace: str, context: str) -> None:
    info(f"Stopping store {store_name}...")

    helm_uninstall(release_name=store_name, context=context, namespace=namespace)

    stop_port_forward(service_name=f"{store_name}-store-service")

    info(f"Store {store_name} stopped successfully.")



"""
Script to populate the server with basic fixtures.
"""

import time
import traceback
from logging import info

from vantage6.common import error

from vantage6.client import Client
from vantage6.client.utils import LogLevel

from vantage6.cli.sandbox.populate.helpers.connect_store import connect_store
from vantage6.cli.sandbox.populate.helpers.delete_fixtures import delete_fixtures
from vantage6.cli.sandbox.populate.helpers.load_fixtures import create_fixtures
from vantage6.cli.sandbox.populate.helpers.utils import NodeConfigCreationDetails


def populate_server_dev(
    server_url: str,
    auth_url: str,
    number_of_nodes: int,
    node_config_creation_details: NodeConfigCreationDetails,
) -> str | None:
    """
    Populate the server with basic fixtures.

    Parameters
    ----------
    server_url : str
        The URL of the server to connect to.
    auth_url : str
        The URL of the auth service to connect to.
    number_of_nodes : int
        The number of nodes to create.
    node_config_creation_details : NodeConfigCreationDetails | None
        The details of the node config creation. If None, the node configs will not be
        created.
    clear_dev_folders : bool
        Whether to clear everything inthe dev directory before creating the new
        resources.
    create_node_configs: bool
        Whether to create the node configs. Default is True.

    Returns
    -------
    str | None
        A report of the creation process, or None if an error occurred.

    Raises
    ------
    Exception
        If connection/authentication to the server fails.
    """
    client = _initalize_client(server_url, auth_url)

    # Create new resources in the server
    # Delete existing resources in the server first, before creating new ones.
    try:
        report_deletion = delete_fixtures(client)
        report_creation = create_fixtures(
            client,
            number_of_nodes,
            node_config_creation_details=node_config_creation_details,
            clear_dev_folders=True,
        )
        report_store = connect_store(client)

        return report_deletion + "\n" + report_creation + "\n" + report_store

    except Exception:
        error("=" * 80)
        error("Failed to populate server")
        error(traceback.format_exc())
        error("=" * 80)


def populate_server_sandbox(
    server_url: str,
    auth_url: str,
    number_of_nodes: int,
) -> dict:
    """
    Populate sandbox server with basic resources.

    Parameters
    ----------
    server_url : str
        The URL of the server to connect to.
    auth_url : str
        The URL of the auth service to connect to.
    number_of_nodes : int
        The number of nodes to create.

    Returns
    -------
    dict
        A dictionary containing the report of the creation process.

    Raises
    ------
    Exception
        If connection/authentication to the server fails.
    """
    client = _initalize_client(server_url, auth_url)

    try:
        delete_fixtures(client)
        report_creation = create_fixtures(
            client,
            number_of_nodes,
            clear_dev_folders=False,
            return_as_dict=True,
        )
        connect_store(client)
    except Exception:
        error("=" * 80)
        error("Failed to populate server")
        error(traceback.format_exc())
        error("=" * 80)
        exit(1)

    # return the details of the created nodes so that config files can be created
    return report_creation["nodes"]["created"]


def _initalize_client(server_url, auth_url) -> Client:
    """
    Initialize an authenticated client to the server.

    The server may not be ready yet, so we retry until it is.

    Parameters
    ----------
    server_url : str
        The URL of the server to connect to.
    auth_url : str
        The URL of the auth service to connect to.

    Returns
    -------
    Client
        An authenticated client to the server.
    """
    client = Client(
        server_url=server_url,
        auth_url=auth_url,
        log_level=LogLevel.ERROR,
    )

    info("Waiting for authentication...")
    max_attempts = 120
    attempt = 1

    while attempt <= max_attempts:
        try:
            print(".", end="", flush=True)
            client.authenticate()
            info("Successfully authenticated with server!")
            break
        except Exception as e:
            if attempt == max_attempts:
                error(
                    f"Failed to authenticate after {max_attempts} attempts. "
                    "Server may not be online."
                )
                raise e

            time.sleep(5)
            attempt += 1

    return client



import os
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from vantage6.common.globals import LOCALHOST, InstanceType, Ports

from vantage6.cli.globals import APPNAME, InfraComponentName
from vantage6.cli.node.attach import cli_node_attach
from vantage6.cli.node.common import create_client_and_authenticate
from vantage6.cli.node.create_private_key import cli_node_create_private_key
from vantage6.cli.node.files import cli_node_files
from vantage6.cli.node.list import cli_node_list
from vantage6.cli.node.new import cli_node_new_configuration
from vantage6.cli.node.restart import cli_node_restart
from vantage6.cli.node.start import cli_node_start
from vantage6.cli.node.stop import cli_node_stop


class NodeCLITest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        return super().setUpClass()

    @patch("vantage6.cli.context.node.NodeContext.available_configurations")
    @patch("vantage6.cli.common.list.find_running_service_names")
    def test_list(self, find_running_service_names, available_configurations):
        """A container list and their current status."""
        # https://docs.python.org/3/library/unittest.mock.html#mock-names-and-the-name-attribute

        node_name = "iknl"
        find_running_service_names.return_value = [f"{APPNAME}-{node_name}-user-node"]

        # returns a list of configurations and failed inports
        def side_effect(system_folders):
            config = MagicMock()
            config.name = node_name
            if not system_folders:
                return [[config], []]
            else:
                return [[config], []]

        available_configurations.side_effect = side_effect

        # invoke CLI method
        runner = CliRunner()
        result = runner.invoke(cli_node_list, [])

        # validate exit code
        self.assertEqual(result.exit_code, 0)

        # check printed lines
        self.assertEqual(
            result.output,
            "\nName                     Status          System/User\n"
            "-----------------------------------------------------\n"
            "iknl                     Not running     System \n"
            "iknl                     Running         User   \n"
            "-----------------------------------------------------\n",
        )

    @patch("vantage6.cli.common.new.select_context_and_namespace")
    @patch("vantage6.cli.common.new.make_configuration")
    @patch("vantage6.cli.common.new.ensure_config_dir_writable")
    @patch("vantage6.cli.node.common.NodeContext")
    def test_new_config(self, context, permissions, make_configuration, ctx_ns):
        """No error produced when creating new configuration."""
        context.config_exists.return_value = False
        permissions.return_value = True
        make_configuration.return_value = "/some/file/path"
        ctx_ns.return_value = ("test-context", "test-namespace")

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some-name",
            ],
        )

        # check that info message is produced
        self.assertEqual(result.output[:7], "[info ]")

        # check OK exit code
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.common.new.make_configuration")
    def test_new_config_replace_whitespace_in_name(self, make_configuration):
        """Whitespaces are replaced in the name."""

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some name",
            ],
        )

        self.assertTrue("[error] - Invalid name: 'some name'." in result.output)
        self.assertEqual(result.exit_code, 1)

    # the context is mocked for select_context_class where it is instantiated
    @patch("vantage6.cli.common.new.select_context_and_namespace")
    @patch("vantage6.cli.context.NodeContext")
    def test_new_config_already_exists(self, context, ctx_ns):
        """No duplicate configurations are allowed."""

        context.config_exists.return_value = True
        ctx_ns.return_value = ("test-context", "test-namespace")

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some-name",
            ],
        )

        # check that error is produced
        self.assertIn(
            "[error] - Configuration some-name already exists!", result.output
        )

        # check non-zero exit code
        self.assertEqual(result.exit_code, 1)

    @patch("vantage6.cli.common.new.select_context_and_namespace")
    @patch("vantage6.cli.common.new.ensure_config_dir_writable")
    @patch("vantage6.cli.node.common.NodeContext")
    def test_new_write_permissions(self, context, permissions, ctx_ns):
        """User needs write permissions."""

        context.config_exists.return_value = False
        permissions.return_value = False
        ctx_ns.return_value = ("test-context", "test-namespace")

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some-name",
            ],
        )

        # check that error is produced
        self.assertTrue("[error] - Your user does not have write" in result.output)

        # check non-zero exit code
        self.assertEqual(result.exit_code, 1)

    @patch("vantage6.cli.common.decorator.get_context")
    @patch("vantage6.cli.common.decorator.select_configuration_questionnaire")
    def test_files(self, select_config, context):
        """No errors produced when retrieving filepaths."""

        context.return_value = MagicMock(
            config_file="/file.yaml", log_file="/log.log", data_dir="/dir"
        )
        context.return_value.databases.items.return_value = [["label", "/file.db"]]
        select_config.return_value = "iknl"

        runner = CliRunner()
        result = runner.invoke(cli_node_files, [])

        # we check that no warnings have been produced
        self.assertEqual(result.output[:7], "[info ]")

        # check status code is OK
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.common.NodeContext")
    def test_files_non_existing_config(self, context):
        """An error is produced when a non existing config is used."""

        context.config_exists.return_value = False

        runner = CliRunner()
        result = runner.invoke(cli_node_files, ["--name", "non-existing"])

        # Check that error is produced
        self.assertEqual(result.output[:7], "[error]")

        # check for non zero exit-code
        self.assertNotEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.start.select_context_and_namespace")
    @patch("os.makedirs")
    @patch("vantage6.cli.common.decorator.get_context")
    @patch("vantage6.cli.node.start.helm_install")
    @patch("vantage6.cli.node.start.start_port_forward")
    def test_start(
        self,
        start_port_forward,
        helm_install,
        context,
        os_makedirs,
        ctx_ns,
    ):
        """Start node without errors"""

        ctx = MagicMock(
            config={"node": {"proxyPort": 8080, "image": "my-node-image"}},
            config_file="/config.yaml",
            data_dir=Path("."),
            log_dir=Path("."),
            helm_release_name="vantage6-test-node",
            is_sandbox=False,
        )
        ctx.config_exists.return_value = True
        ctx.name = "test-node"
        context.return_value = ctx
        ctx_ns.return_value = ("test-context", "test-namespace")

        runner = CliRunner()
        result = runner.invoke(cli_node_start, ["--name", "test-node"])

        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.common.stop.select_context_and_namespace")
    @patch("vantage6.cli.common.stop.get_context")
    @patch("vantage6.cli.common.utils.find_running_service_names")
    @patch("vantage6.cli.node.stop._stop_node")
    def test_stop(self, _stop_node, find_running_service_names, get_context, ctx_ns):
        node_name = "iknl"
        node_helm_name = f"{APPNAME}-{node_name}-user-node"
        find_running_service_names.return_value = [node_helm_name]
        get_context.return_value = MagicMock(helm_release_name=node_helm_name)
        ctx_ns.return_value = ("test-context", "test-namespace")

        runner = CliRunner()
        result = runner.invoke(cli_node_stop, ["--name", "iknl"])

        self.assertEqual(result.exit_code, 0)
        self.assertIsNone(result.exception)

    @patch("vantage6.cli.common.stop.select_context_and_namespace")
    @patch("vantage6.cli.node.restart.select_context_and_namespace")
    @patch("vantage6.cli.node.restart.subprocess.run")
    @patch("vantage6.cli.common.stop.get_context")
    @patch("vantage6.cli.node.stop._stop_node")
    def test_restart(
        self, _stop_node, get_context, subprocess_run, ctx_ns_stop, ctx_ns_restart
    ):
        """Restart a node without errors."""
        node_name = "iknl"
        node_helm_name = f"{APPNAME}-{node_name}-user-node"
        get_context.return_value = MagicMock(helm_release_name=node_helm_name)
        ctx_ns_stop.return_value = ("test-context", "test-namespace")
        ctx_ns_restart.return_value = ("test-context", "test-namespace")

        # Mock subprocess.run to handle both helm calls and node start calls
        def mock_subprocess_run(*args, **kwargs):
            # Check if this is a helm command (contains 'helm list')
            if "helm" in args[0] and "list" in args[0]:
                return MagicMock(
                    stdout='[{"name": "' + node_helm_name + '", "status": "deployed"}]',
                    returncode=0,
                )
            else:
                # For other subprocess calls (like starting the node)
                return MagicMock(returncode=0)

        subprocess_run.side_effect = mock_subprocess_run
        runner = CliRunner()
        result = runner.invoke(cli_node_restart, ["--name", "iknl"])
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.attach.attach_logs")
    def test_attach(self, attach_logs):
        """Attach pod logs without errors."""
        runner = CliRunner()
        runner.invoke(cli_node_attach)
        attach_logs.assert_called_once_with(
            name=None,
            instance_type=InstanceType.NODE,
            infra_component=InfraComponentName.NODE,
            system_folders=False,
            context=None,
            namespace=None,
            is_sandbox=False,
        )

    @patch("vantage6.cli.node.create_private_key.create_client_and_authenticate")
    @patch("vantage6.cli.node.common.NodeContext")
    @patch("vantage6.cli.node.create_private_key.NodeContext")
    def test_create_private_key(self, context, common_context, client):
        common_context.config_exists.return_value = True
        context.return_value.type_data_folder.return_value = Path(".")
        client.return_value = MagicMock(whoami=MagicMock(organization_name="Test"))
        # client.whoami.organization_name = "Test"
        runner = CliRunner()

        result = runner.invoke(cli_node_create_private_key, ["--name", "application"])

        self.assertEqual(result.exit_code, 0)

        # remove the private key file again
        os.remove("privkey_Test.pem")

    @patch("vantage6.cli.node.create_private_key.RSACryptor")
    @patch("vantage6.cli.node.create_private_key.create_client_and_authenticate")
    @patch("vantage6.cli.node.common.NodeContext")
    @patch("vantage6.cli.node.create_private_key.NodeContext")
    def test_create_private_key_overwite(
        self, context, common_context, client, cryptor
    ):
        common_context.config_exists.return_value = True
        context.return_value.type_data_folder.return_value = Path(".")
        client.return_value = MagicMock(whoami=MagicMock(organization_name="Test"))
        cryptor.create_public_key_bytes.return_value = b""
        # client.whoami.organization_name = "Test"

        runner = CliRunner()

        # overwrite
        with runner.isolated_filesystem():
            with open("privkey_iknl.pem", "w") as f:
                f.write("does-not-matter")

            result = runner.invoke(
                cli_node_create_private_key,
                ["--name", "application", "--overwrite", "--organization-name", "iknl"],
            )
        self.assertEqual(result.exit_code, 0)

        # do not overwrite
        with runner.isolated_filesystem():
            with open("privkey_iknl.pem", "w") as f:
                f.write("does-not-matter")

            result = runner.invoke(
                cli_node_create_private_key,
                ["--name", "application", "--organization-name", "iknl"],
            )

        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.common.NodeContext")
    def test_create_private_key_config_not_found(self, context):
        context.config_exists.return_value = False

        runner = CliRunner()
        result = runner.invoke(cli_node_create_private_key, ["--name", "application"])

        self.assertEqual(result.exit_code, 1)

    @patch("vantage6.cli.node.common.info")
    @patch("vantage6.cli.node.common.debug")
    @patch("vantage6.cli.node.common.error")
    @patch("vantage6.cli.node.common.UserClient")
    def test_client(self, client, error, debug, info):
        ctx = MagicMock(
            config={
                "node": {
                    "server": {
                        "url": LOCALHOST,
                        "port": Ports.DEV_SERVER.value,
                        "path": "",
                    }
                }
            }
        )

        # should not trigger an exception
        try:
            create_client_and_authenticate(ctx)
        except Exception:
            self.fail("Raised an exception!")

        # client raises exception
        client.side_effect = Exception("Boom!")
        with self.assertRaises(Exception):
            create_client_and_authenticate(ctx)



from pathlib import Path

import click
from colorama import Fore, Style

from vantage6.common import error
from vantage6.common.globals import Ports

from vantage6.cli.common.utils import select_context_and_namespace
from vantage6.cli.context.server import ServerContext
from vantage6.cli.sandbox.config.core import CoreSandboxConfigManager
from vantage6.cli.sandbox.start import execute_sandbox_start
from vantage6.cli.server.common import get_server_context
from vantage6.cli.utils import prompt_config_name


@click.command()
@click.option(
    "-n", "--name", default=None, type=str, help="Name for your development setup"
)
@click.option(
    "--num-nodes",
    type=int,
    default=3,
    help="Generate this number of nodes in the development network",
)
@click.option(
    "-p",
    "--server-port",
    type=int,
    default=Ports.DEV_SERVER.value,
    help=f"Port to run the server on. Default is {Ports.DEV_SERVER}.",
)
@click.option(
    "--ui-port",
    type=int,
    default=Ports.DEV_UI.value,
    help=f"Port to run the UI on. Default is {Ports.DEV_UI}.",
)
@click.option(
    "--algorithm-store-port",
    type=int,
    default=Ports.DEV_ALGO_STORE.value,
    help=f"Port to run the algorithm store on. Default is {Ports.DEV_ALGO_STORE}.",
)
@click.option(
    "--server-image",
    type=str,
    default=None,
    help="Server docker image to use when setting up resources for "
    "the development server",
)
@click.option(
    "--ui-image",
    type=str,
    default=None,
    help="UI docker image to specify in configuration files. Will be used on startup of"
    " the network",
)
@click.option(
    "--store-image",
    type=str,
    default=None,
    help="Algorithm store docker image to use when setting up resources for "
    "the development algorithm store",
)
@click.option(
    "--node-image",
    type=str,
    default=None,
    help="Node docker image to use when setting up resources for the development node",
)
@click.option(
    "--extra-server-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional server "
    "configuration. This will be appended to the server "
    "configuration file",
)
@click.option(
    "--extra-node-config",
    type=click.Path("rb"),
    default=None,
    help="YAML File with additional node configuration. This will be"
    " appended to each of the node configuration files",
)
@click.option(
    "--extra-store-config",
    type=click.Path("rb"),
    default=None,
    help="YAML File with additional algorithm store configuration. This will be"
    " appended to the algorithm store configuration file",
)
@click.option(
    "--extra-auth-config",
    type=click.Path("rb"),
    default=None,
    help="YAML File with additional auth configuration. This will be"
    " appended to the auth configuration file",
)
@click.option(
    "--add-dataset",
    type=(str, click.Path()),
    default=None,
    multiple=True,
    help="Add a dataset to the nodes. The first argument is the label of the database, "
    "the second is the path to the dataset file.",
)
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--k8s-node-name", default="docker-desktop", help="Kubernetes node name to use"
)
@click.option(
    "--data-dir",
    type=click.Path(exists=True),
    default=None,
    help="Path to a custom data directory to use. This option is especially useful "
    "on WSL because of mount issues for default directories",
)
@click.option(
    "--local-chart-dir",
    type=click.Path(exists=True),
    default=None,
    help="Local chart repository to use.",
)
@click.pass_context
def cli_new_sandbox(
    click_ctx: click.Context,
    name: str,
    num_nodes: int,
    server_port: int,
    ui_port: int,
    algorithm_store_port: int,
    server_image: str | None,
    ui_image: str | None,
    store_image: str | None,
    node_image: str | None,
    extra_server_config: Path | None,
    extra_node_config: Path | None,
    extra_store_config: Path | None,
    extra_auth_config: Path | None,
    add_dataset: tuple[str, Path] | None,
    context: str | None,
    namespace: str | None,
    k8s_node_name: str,
    data_dir: str | None,
    local_chart_dir: Path | None,
) -> None:
    """
    Create a sandbox environment.
    """

    # Prompt for the k8s namespace and context
    context, namespace = select_context_and_namespace(
        context=context,
        namespace=namespace,
    )

    if data_dir is not None:
        data_dir = Path(data_dir)
        if not data_dir.exists():
            error(f"Data directory {data_dir} does not exist!")
            exit(1)

    server_name = prompt_config_name(name)
    if ServerContext.config_exists(server_name, False, is_sandbox=True):
        error(f"Configuration {Fore.RED}{server_name}{Style.RESET_ALL} already exists!")
        exit(1)

    sb_config_manager = CoreSandboxConfigManager(
        server_name=server_name,
        server_port=server_port,
        ui_port=ui_port,
        algorithm_store_port=algorithm_store_port,
        server_image=server_image,
        ui_image=ui_image,
        store_image=store_image,
        extra_server_config=extra_server_config,
        extra_store_config=extra_store_config,
        extra_auth_config=extra_auth_config,
        context=context,
        namespace=namespace,
        k8s_node_name=k8s_node_name,
        custom_data_dir=data_dir,
    )

    sb_config_manager.generate_server_configs()

    ctx = get_server_context(server_name, False, ServerContext, is_sandbox=True)

    execute_sandbox_start(
        click_ctx=click_ctx,
        ctx=ctx,
        server_name=server_name,
        context=context,
        namespace=namespace,
        num_nodes=num_nodes,
        initialize=True,
        node_image=node_image,
        k8s_node_name=k8s_node_name,
        extra_node_config=extra_node_config,
        add_dataset=add_dataset,
        custom_data_dir=data_dir,
        local_chart_dir=local_chart_dir,
    )



import json
import subprocess
from pathlib import Path

import questionary as q

from vantage6.common import error
from vantage6.common.globals import (
    APPNAME,
    SANDBOX_SUFFIX,
    InstanceType,
)

from vantage6.cli.config import CliConfig
from vantage6.cli.globals import CLICommandName
from vantage6.cli.utils import validate_input_cmd_args


def select_context_and_namespace(
    context: str | None = None,
    namespace: str | None = None,
) -> tuple[str, str]:
    """
    Select the context and namespace to use.

    This uses the CLI config to compare the provided context and namespace with the
    last used context and namespace. If the provided context and namespace are not
    the same as the last used context and namespace, the CLI config is updated.

    Parameters
    ----------
    context : str, optional
        The Kubernetes context to use.
    namespace : str, optional
        The Kubernetes namespace to use.

    Returns
    -------
    tuple[str, str]
        The context and namespace to use
    """
    cli_config = CliConfig()

    return cli_config.compare_changes_config(
        context=context,
        namespace=namespace,
    )


def create_directory_if_not_exists(directory: Path) -> None:
    """
    Create a directory.
    """
    try:
        directory.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        error(f"Failed to create directory {directory}: {e}")
        exit(1)


def find_running_service_names(
    instance_type: InstanceType,
    only_system_folders: bool = False,
    only_user_folders: bool = False,
    context: str | None = None,
    namespace: str | None = None,
    sandbox: bool = False,
) -> list[str]:
    """
    List running Vantage6 servers.

    Parameters
    ----------
    instance_type : InstanceType
        The type of instance to find running services for
    only_system_folders : bool, optional
        Whether to look for system-based services or not. By default False.
    only_user_folders : bool, optional
        Whether to look for user-based services or not. By default False.
    context : str, optional
        The Kubernetes context to use.
    namespace : str, optional
        The Kubernetes namespace to use.
    sandbox : bool, optional
        Whether to look for sandbox services or not. By default False.

    Returns
    -------
    list[str]
        List of release names that are running
    """
    # Input validation
    validate_input_cmd_args(context, "context name", allow_none=True)
    validate_input_cmd_args(namespace, "namespace name", allow_none=True)
    validate_input_cmd_args(instance_type, "instance type", allow_none=False)
    if only_system_folders and only_user_folders:
        error("Cannot use both only_system_folders and only_user_folders")
        exit(1)

    # Create the command
    command = [
        "helm",
        "list",
        "--output",
        "json",  # Get structured output
    ]

    if context:
        command.extend(["--kube-context", context])

    if namespace:
        command.extend(["--namespace", namespace])
    else:
        command.extend(["--all-namespaces"])

    try:
        result = subprocess.run(
            command,
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        error(f"Failed to list Helm releases: {e}")
        return []
    except FileNotFoundError:
        error(
            "Helm command not found. Please ensure Helm is installed and available in "
            "the PATH."
        )
        return []

    try:
        releases = json.loads(result.stdout)
    except json.JSONDecodeError:
        error("Failed to parse Helm output as JSON")
        return []

    # filter services for the vantage6 services that are sought. These have
    # the following pattern:
    # f"{APPNAME}-{name}-{scope}-{instance_type.value}"

    # filter for the instance type
    svc_starts_with = f"{APPNAME}-"
    if only_system_folders:
        svc_ends_with = f"system-{instance_type.value}"
    elif only_user_folders:
        svc_ends_with = f"user-{instance_type.value}"
    else:
        svc_ends_with = f"-{instance_type.value}"

    matching_services = []
    for release in releases:
        release_name = release.get("name", "")

        # Check if this is a Vantage6 server release
        is_matching_service = (
            release_name.startswith(svc_starts_with)
            and release_name.endswith(svc_ends_with)
            and instance_type.value in release_name
        )

        if is_matching_service:
            matching_services.append(release_name)

    return matching_services


def select_running_service(
    running_services: list[str],
    instance_type: InstanceType,
) -> str:
    """
    Select a running service from the list of running services.
    """
    try:
        name = q.select(
            f"Select a {instance_type.value}:",
            choices=running_services,
        ).unsafe_ask()
    except KeyboardInterrupt:
        error("Aborted by user!")
        exit(1)
    return name


def get_config_name_from_service_name(service_name: str) -> str:
    """
    Get the config name from a service name.
    """
    # helm release name is structured as:
    # f"{APPNAME}-{name}-{scope}-{instance_type}"
    # we want to get the name from the service name
    return "-".join(service_name.split("-")[1:-2])


def get_main_cli_command_name(instance_type: InstanceType) -> str:
    """
    Get the main CLI command name for a given instance type.

    Parameters
    ----------
    instance_type : InstanceType
        The type of instance to get the main CLI command name for
    """
    if instance_type == InstanceType.SERVER:
        return CLICommandName.SERVER.value
    elif instance_type == InstanceType.ALGORITHM_STORE:
        return CLICommandName.ALGORITHM_STORE.value
    elif instance_type == InstanceType.NODE:
        return CLICommandName.NODE.value
    elif instance_type == InstanceType.AUTH:
        return CLICommandName.AUTH.value
    else:
        raise ValueError(f"Invalid instance type: {instance_type}")


def check_running(
    helm_release_name: str, instance_type: InstanceType, name: str, system_folders: bool
) -> bool:
    """
    Check if the instance is already running.

    Parameters
    ----------
    helm_release_name : str
        The name of the Helm release.
    instance_type : InstanceType
        The type of instance to check
    name : str
        The name of the instance to check
    system_folders : bool
        Whether to use system folders or not

    Returns
    -------
    bool
        True if the instance is already running, False otherwise
    """
    running_services = find_running_service_names(
        instance_type=instance_type,
        only_system_folders=system_folders,
        only_user_folders=not system_folders,
    )
    return helm_release_name in running_services


def get_config_name_from_helm_release_name(
    helm_release_name: str, is_store: bool = False
) -> str:
    """
    Get the config name from a helm release name.

    Parameters
    ----------
    helm_release_name : str
        The name of the Helm release
    is_store : bool, optional
        Whether the instance is a store or not. By default False.

    Returns
    -------
    str
        The config name
    """
    # helm release name is structured as:
    # f"{APPNAME}-{name}-{scope}-{instance_type}"
    # we want to get the name from the service name
    if is_store:
        # for store, the instance type is `algorithm-store` which contains an additional
        # hyphen
        return "-".join(helm_release_name.split("-")[1:-3])
    else:
        return "-".join(helm_release_name.split("-")[1:-2])


def extract_name_and_is_sandbox(name: str | None, is_sandbox: bool) -> tuple[str, bool]:
    """
    Extract the name and is_sandbox from the name.

    Note that the name may be None: this occurs before when this function is called
    before the user has selected a name. This scenario is fine because when the user
    selects a name interactively, the name never ends with the .sandbox suffix.

    Parameters
    ----------
    name : str | None
        The name of the instance
    is_sandbox : bool
        Whether the instance is a sandbox instance

    Returns
    -------
    tuple[str, bool]
        The name and is_sandbox
    """
    if name and name.endswith(SANDBOX_SUFFIX):
        return name[: -len(SANDBOX_SUFFIX)], True
    else:
        return name, is_sandbox



from pathlib import Path

from vantage6.common.globals import HTTP_LOCALHOST, InstanceType, Ports

from vantage6.cli.common.new import new
from vantage6.cli.sandbox.config.base import BaseSandboxConfigManager


class CoreSandboxConfigManager(BaseSandboxConfigManager):
    """
    Class to store the sandbox configurations.

    Parameters
    ----------
    server_name : str
        Name of the server.
    server_port : int
        Port of the server.
    ui_port : int
        Port of the UI.
    algorithm_store_port : int
        Port of the algorithm store.
    server_image : str | None
        Image of the server.
    store_image : str | None
        Image of the algorithm store.
    ui_image : str | None
        Image of the UI.
    extra_server_config : Path | None
        Path to the extra server configuration file.
    extra_store_config : Path | None
        Path to the extra algorithm store configuration file.
    context : str | None
        Kubernetes context.
    namespace : str | None
        Kubernetes namespace.
    k8s_node_name : str
        Kubernetes node name.
    custom_data_dir : Path | None
        Path to the custom data directory. Useful on WSL because of mount issues for
        default directories.
    """

    def __init__(
        self,
        server_name: str,
        server_port: int,
        ui_port: int,
        algorithm_store_port: int,
        server_image: str | None,
        store_image: str | None,
        ui_image: str | None,
        extra_server_config: Path | None,
        extra_store_config: Path | None,
        extra_auth_config: Path | None,
        context: str,
        namespace: str,
        k8s_node_name: str,
        custom_data_dir: Path | None = None,
    ) -> None:
        super().__init__(server_name, custom_data_dir)

        self.server_port = server_port
        self.ui_port = ui_port
        self.algorithm_store_port = algorithm_store_port
        self.server_image = server_image
        self.store_image = store_image
        self.ui_image = ui_image
        self.extra_server_config = extra_server_config
        self.extra_store_config = extra_store_config
        self.extra_auth_config = extra_auth_config
        self.context = context
        self.namespace = namespace

        self.server_config_file = None
        self.store_config_file = None
        self.auth_config_file = None
        self.k8s_node_name = k8s_node_name

    def generate_server_configs(self) -> None:
        """Generates the demo network."""

        self._create_auth_config()

        self._create_vserver_config()

        self._create_algo_store_config()

    def __server_config_return_func(self, extra_config: dict, data_dir: Path) -> dict:
        """
        Return a dict with server configuration values to be used in creating the
        config file.

        Parameters
        ----------
        extra_config : dict
            Extra configuration (parsed from YAML) to be added to the server
            configuration.
        data_dir : Path
            Path to the data directory.

        Returns
        -------
        dict
            Dictionary with server configuration values.
        """
        store_service = (
            f"vantage6-{self.server_name}-store-user-algorithm-store-store-service"
        )
        store_address = (
            f"http://{store_service}.{self.namespace}.svc.cluster.local:"
            f"{Ports.DEV_ALGO_STORE}"
        )
        config = {
            "server": {
                "baseUrl": f"{HTTP_LOCALHOST}:{self.server_port}",
                # TODO: v5+ set to latest v5 image
                # TODO make this configurable
                "image": (
                    self.server_image
                    or "harbor2.vantage6.ai/infrastructure/server:5.0.0a37"
                ),
                "algorithm_stores": [
                    {
                        "name": "Local store",
                        "url": store_address,
                        "api_path": "/store",
                    }
                ],
                "logging": {
                    "level": "DEBUG",
                },
                "jwt": {
                    "secret": "development-constant-secret!",
                },
                "dev": {
                    "host_uri": (
                        "host.docker.internal"
                        if self.k8s_node_name == "docker-desktop"
                        else "172.17.0.1"
                    ),
                    "store_address": store_address,
                },
                "keycloakUrl": (
                    f"http://vantage6-{self.server_name}-auth-user-auth-keycloak."
                    f"{self.namespace}.svc.cluster.local"
                ),
            },
            "rabbitmq": {},
            "database": {
                "volumePath": str(data_dir),
                "k8sNodeName": self.k8s_node_name,
            },
            "ui": {
                "port": self.ui_port,
                # TODO: v5+ set to latest v5 image
                # TODO: make this configurable
                "image": (
                    self.ui_image or "harbor2.vantage6.ai/infrastructure/ui:5.0.0a37"
                ),
            },
        }

        # merge the extra config with the server config
        if extra_config is not None:
            config.update(extra_config)

        return config

    def _create_vserver_config(self) -> None:
        """Creates server configuration file (YAML)."""

        data_dir = self._create_and_get_data_dir(instance_type=InstanceType.SERVER)

        extra_config = self._read_extra_config_file(self.extra_server_config)
        if self.ui_image is not None:
            ui_config = extra_config.get("ui", {}) if extra_config is not None else {}
            ui_config["image"] = self.ui_image
            extra_config["ui"] = ui_config

        # Create the server config file
        self.server_config_file = new(
            config_producing_func=self.__server_config_return_func,
            config_producing_func_args=(extra_config, data_dir),
            name=self.server_name,
            system_folders=False,
            namespace=self.namespace,
            context=self.context,
            type_=InstanceType.SERVER,
            is_sandbox=True,
        )

    def _create_algo_store_config(self) -> None:
        """Create algorithm store configuration file (YAML)."""

        extra_config = self._read_extra_config_file(self.extra_store_config)

        data_dir = self._create_and_get_data_dir(InstanceType.ALGORITHM_STORE)

        self.store_config_file = new(
            config_producing_func=self.__algo_store_config_return_func,
            config_producing_func_args=(extra_config, data_dir),
            name=f"{self.server_name}-store",
            system_folders=False,
            namespace=self.namespace,
            context=self.context,
            type_=InstanceType.ALGORITHM_STORE,
            is_sandbox=True,
        )

    def __algo_store_config_return_func(
        self, extra_config: dict, data_dir: Path
    ) -> dict:
        """
        Return a dict with algorithm store configuration values to be used in creating
        the config file.

        Returns
        -------
        dict
            Dictionary with algorithm store configuration values.
        """
        config = {
            "store": {
                "internal": {
                    "port": self.algorithm_store_port,
                },
                "logging": {
                    "level": "DEBUG",
                },
                "vantage6ServerUri": f"{HTTP_LOCALHOST}:{self.server_port}",
                "image": (
                    self.store_image
                    or "harbor2.vantage6.ai/infrastructure/algorithm-store:5.0.0a37"
                ),
                "keycloakUrl": (
                    f"http://vantage6-{self.server_name}-auth-user-auth-keycloak."
                    f"{self.namespace}.svc.cluster.local"
                ),
                "policies": {
                    "allowLocalhost": True,
                    "assignReviewOwnAlgorithm": True,
                },
                "dev": {
                    "host_uri": (
                        "host.docker.internal"
                        if self.k8s_node_name == "docker-desktop"
                        else "172.17.0.1"
                    ),
                    "disable_review": True,
                    "review_own_algorithm": True,
                },
            },
            "database": {
                "volumePath": str(data_dir),
                "k8sNodeName": self.k8s_node_name,
            },
        }

        # merge the extra config with the algorithm store config
        if extra_config is not None:
            config.update(extra_config)

        return config

    def _create_auth_config(self) -> None:
        """Create auth configuration file (YAML)."""
        self.auth_config_file = new(
            config_producing_func=self.__auth_config_return_func,
            config_producing_func_args=(self.extra_auth_config,),
            name=f"{self.server_name}-auth",
            system_folders=False,
            namespace=self.namespace,
            context=self.context,
            type_=InstanceType.AUTH,
            is_sandbox=True,
        )

    def __auth_config_return_func(self, extra_config: dict) -> dict:
        """
        Return a dict with auth configuration values to be used in creating the
        config file.
        """

        config = {
            "keycloak": {
                "production": False,
                "no_password_update_required": True,
                "redirectUris": [
                    f"{HTTP_LOCALHOST}:7600",
                    f"{HTTP_LOCALHOST}:7681",
                ],
            },
        }

        # merge the extra config with the auth config
        if extra_config is not None:
            config.update(extra_config)

        return config



from __future__ import annotations

from vantage6.common.context import AppContext
from vantage6.common.globals import InstanceType

from vantage6.cli import __version__
from vantage6.cli.configuration_manager import AuthConfigurationManager
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS as S_FOL


class AuthContext(AppContext):
    """
    Context class for the keycloak authentication server.

    Parameters
    ----------
    instance_name : str
        Name of the configuration instance, corresponds to the filename
        of the configuration file.
    system_folders : bool, optional
        System wide or user configuration, by default S_FOL
    is_sandbox : bool, optional
        Whether the configuration is a sandbox configuration, by default False
    """

    # The auth configuration manager is aware of the structure of the auth
    # configuration file and makes sure only valid configuration can be loaded.
    INST_CONFIG_MANAGER = AuthConfigurationManager

    def __init__(
        self,
        instance_name: str,
        system_folders: bool = S_FOL,
        is_sandbox: bool = False,
    ):
        super().__init__(
            InstanceType.AUTH,
            instance_name,
            system_folders=system_folders,
            is_sandbox=is_sandbox,
        )
        self.log.info("vantage6 version '%s'", __version__)

    @classmethod
    def from_external_config_file(
        cls, path: str, system_folders: bool = S_FOL
    ) -> AuthContext:
        """
        Create a server context from an external configuration file. External
        means that the configuration file is not located in the default folders
        but its location is specified by the user.

        Parameters
        ----------
        path : str
            Path of the configuration file
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL

        Returns
        -------
        ServerContext
            Server context object
        """
        return super().from_external_config_file(
            path,
            InstanceType.AUTH,
            system_folders,
        )

    @classmethod
    def config_exists(
        cls,
        instance_name: str,
        system_folders: bool = S_FOL,
        is_sandbox: bool = False,
    ) -> bool:
        """
        Check if a configuration file exists.

        Parameters
        ----------
        instance_name : str
            Name of the configuration instance, corresponds to the filename
            of the configuration file.
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL
        is_sandbox : bool, optional
            Whether the configuration is a sandbox configuration, by default False

        Returns
        -------
        bool
            Whether the configuration file exists or not
        """
        return super().base_config_exists(
            InstanceType.AUTH,
            instance_name,
            system_folders=system_folders,
            is_sandbox=is_sandbox,
        )

    @classmethod
    def available_configurations(
        cls, system_folders: bool = S_FOL, is_sandbox: bool = False
    ) -> tuple[list, list]:
        """
        Find all available auth configurations in the default folders.

        Parameters
        ----------
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL
        is_sandbox : bool, optional
            Whether the configuration is a sandbox configuration, by default False

        Returns
        -------
        tuple[list, list]
            The first list contains validated configuration files, the second
            list contains invalid configuration files.
        """
        return super().available_configurations(
            InstanceType.AUTH, system_folders, is_sandbox
        )



from typing import Any

import click
import questionary as q

from vantage6.common.globals import (
    InstanceType,
)

from vantage6.cli.common.new import new
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS


@click.command()
@click.option(
    "-n", "--name", default=None, help="name of the configuration you want to use."
)
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Use system folders instead of user folders. This is the default",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=DEFAULT_SERVER_SYSTEM_FOLDERS,
    help="Use user folders instead of system folders",
)
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option(
    "--namespace",
    default=None,
    help="Kubernetes namespace to use",
)
def cli_auth_new(
    name: str,
    system_folders: bool,
    namespace: str,
    context: str,
) -> None:
    """
    Create a new server configuration.
    """
    new(
        config_producing_func=auth_configuration_questionaire,
        config_producing_func_args=tuple(),
        name=name,
        system_folders=system_folders,
        namespace=namespace,
        context=context,
        type_=InstanceType.AUTH,
    )


def auth_configuration_questionaire() -> dict[str, Any]:
    """
    Kubernetes-specific questionnaire to generate Helm values for the Keycloak helm
    chart.
    """
    config = {"keycloak": {}}

    is_production = q.confirm(
        "Do you want to use production settings? If not, the service will be configured"
        " to be more suitable for development or testing purposes.",
        default=True,
    ).unsafe_ask()

    config["keycloak"]["production"] = is_production

    if is_production:
        ui_url = q.text(
            "Please provide the URL of the UI. This is the URL that users will use to "
            "log in to the service.",
            default="https://ui.vantage6.ai",
        ).unsafe_ask()
        # add http://localhost:7681 as that is used by the Python client
        config["keycloak"]["redirectUris"] = [ui_url, "http://localhost:7681"]

    return config



from collections.abc import Callable
from pathlib import Path

import questionary as q

from vantage6.common import error, info
from vantage6.common.context import AppContext
from vantage6.common.globals import (
    DEFAULT_API_PATH,
    InstanceType,
    Ports,
)

from vantage6.cli.configuration_manager import (
    AlgorithmStoreConfigurationManager,
    AuthConfigurationManager,
    NodeConfigurationManager,
    ServerConfigurationManager,
)
from vantage6.cli.context import select_context_class


def add_common_server_config(
    config: dict, instance_type: InstanceType, instance_name: str
) -> dict:
    """
    Part of the questionaire that is common to all server types (vantage6
    server and algorithm store server).

    Parameters
    ----------
    instance_type : InstanceType
        Type of server instance.
    instance_name : str
        Name of the server instance.

    Returns
    -------
    dict
        Dictionary with new (partial) server configuration
    """
    backend_config = (
        config["server"] if instance_type == InstanceType.SERVER else config["store"]
    )

    backend_config["port"] = q.text(
        "Enter port to which the server listens:",
        default=(
            str(Ports.DEV_SERVER)
            if instance_type == InstanceType.SERVER
            else str(Ports.DEV_ALGO_STORE)
        ),
    ).unsafe_ask()

    res = q.select(
        "Which level of logging would you like?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    ).unsafe_ask()

    backend_config["logging"] = {
        "level": res,
        "file": f"{instance_name}.log",
        "use_console": True,
        "backup_count": 5,
        "max_size": 1024,
        "format": "%(asctime)s - %(name)-14s - %(levelname)-8s - %(message)s",
        "datefmt": "%Y-%m-%d %H:%M:%S",
        "loggers": [
            {"name": "urllib3", "level": "warning"},
            {"name": "socketIO-client", "level": "warning"},
            {"name": "socketio.server", "level": "warning"},
            {"name": "engineio.server", "level": "warning"},
            {"name": "sqlalchemy.engine", "level": "warning"},
        ],
    }

    backend_config["api_path"] = DEFAULT_API_PATH

    # === Database settings ===
    config["database"]["volumePath"] = q.text(
        "Where is your server database located on the host machine?",
        default=f"{Path.cwd()}/dev/.db/db_pv_server",
    ).unsafe_ask()

    config["database"]["k8sNodeName"] = q.text(
        "What is the name of the k8s node where the databases are running?",
        default="docker-desktop",
    ).unsafe_ask()

    is_production = q.confirm(
        "Do you want to use production settings for this server? If not, the server will "
        "be configured to be more suitable for development or testing purposes.",
        default=True,
    ).unsafe_ask()

    if is_production:
        config = _add_production_server_config(config)

    return config, is_production


def _add_production_server_config(config: dict) -> dict:
    """
    Add the production server configuration to the config

    Parameters
    ----------
    config : dict
        The config to add the production server configuration to

    Returns
    -------
    dict
        The config with the production server configuration added
    """
    info("For production environments, it is recommended to use an external database.")
    info("Please provide the URI of the external database.")
    info("Example: postgresql://username:password@localhost:5432/vantage6")

    config["database"]["external"] = True
    config["database"]["uri"] = q.text(
        "Database URI:",
        default="postgresql://vantage6:vantage6@localhost:5432/vantage6",
    ).unsafe_ask()

    return config


def make_configuration(
    config_producing_func: Callable,
    config_producing_func_args: tuple,
    type_: InstanceType,
    instance_name: str,
    system_folders: bool,
    is_sandbox: bool = False,
) -> Path:
    """
    Create a configuration file for a node or server instance.

    Parameters
    ----------
    config_producing_func : Callable
        Function to generate the configuration
    config_producing_func_args : tuple
        Arguments to pass to the config producing function
    type_ : InstanceType
        Type of the instance to create a configuration for
    instance_name : str
        Name of the instance
    system_folders : bool
        Whether to use the system folders or not
    is_sandbox : bool
        Whether to create a sandbox configuration or not

    Returns
    -------
    Path
        Path to the configuration file
    """
    # for defaults and where to save the config
    dirs = AppContext.instance_folders(type_, instance_name, system_folders)

    # invoke function to create configuration file. Usually this is a questionaire
    # but it can also be a function that immediately returns a dict with the
    # configuration.
    config = config_producing_func(*config_producing_func_args)

    # in the case of an environment we need to add it to the current
    # configuration. In the case of application we can simply overwrite this
    # key (although there might be environments present)
    config_file = Path(dirs.get("config")) / (instance_name + ".yaml")
    if type_ == InstanceType.NODE:
        conf_manager = NodeConfigurationManager
    elif type_ == InstanceType.SERVER:
        conf_manager = ServerConfigurationManager
    elif type_ == InstanceType.ALGORITHM_STORE:
        conf_manager = AlgorithmStoreConfigurationManager
    elif type_ == InstanceType.AUTH:
        conf_manager = AuthConfigurationManager
    else:
        raise ValueError(f"Invalid instance type: {type_}")

    if Path(config_file).exists():
        config_manager = conf_manager.from_file(config_file, is_sandbox=is_sandbox)
    else:
        config_manager = conf_manager(instance_name, is_sandbox=is_sandbox)

    config_manager.put(config)
    config_file = config_manager.save(config_file)

    return config_file


def select_configuration_questionnaire(
    type_: InstanceType, system_folders: bool, is_sandbox: bool = False
) -> str:
    """
    Ask which configuration the user wants to use. It shows only configurations
    that are in the default folder.

    Parameters
    ----------
    type_ : InstanceType
        Type of the instance to create a configuration for
    system_folders : bool
        Whether to use the system folders or not
    is_sandbox : bool
        Whether to show only the sandbox configurations or not

    Returns
    -------
    str
        Name of the configuration
    """
    context = select_context_class(type_)
    configs, _ = context.available_configurations(system_folders, is_sandbox)

    # each collection (file) can contain multiple configs. (e.g. test,
    # dev)
    choices = []
    for config_collection in configs:
        choices.append(
            q.Choice(title=f"{config_collection.name:25}", value=config_collection.name)
        )

    if not choices:
        raise Exception("No configurations could be found!")

    # pop the question
    try:
        return q.select(
            "Select the configuration you want to use:", choices=choices
        ).unsafe_ask()
    except KeyboardInterrupt:
        error("Aborted by user!")
        exit(1)



import click

from vantage6.common import info
from vantage6.common.globals import InstanceType

from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.context.auth import AuthContext


@click.command()
@click_insert_context(type_=InstanceType.AUTH)
def cli_auth_files(ctx: AuthContext) -> None:
    """
    List files that belong to a particular auth instance.
    """
    info(f"Configuration file = {ctx.config_file}")



import click

from vantage6.common import info, warning
from vantage6.common.globals import (
    DEFAULT_DOCKER_REGISTRY,
    DEFAULT_NODE_IMAGE,
    DEFAULT_NODE_IMAGE_WO_TAG,
    InstanceType,
)

from vantage6.cli import __version__
from vantage6.cli.common.attach import attach_logs
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.start import (
    helm_install,
    prestart_checks,
    start_port_forward,
)
from vantage6.cli.common.utils import (
    create_directory_if_not_exists,
    select_context_and_namespace,
)
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import ChartName, InfraComponentName
from vantage6.cli.node.common import create_client

from vantage6.node.globals import DEFAULT_PROXY_SERVER_PORT


@click.command()
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--attach/--detach",
    default=False,
    help="Show node logs on the current console after starting the node",
)
@click.option("--local-chart-dir", default=None, help="Local chart directory to use")
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click_insert_context(
    InstanceType.NODE,
    include_name=True,
    include_system_folders=True,
    sandbox_param="sandbox",
)
def cli_node_start(
    ctx: NodeContext,
    name: str,
    system_folders: bool,
    context: str,
    namespace: str,
    attach: bool,
    local_chart_dir: str,
) -> None:
    """
    Start the node.
    """
    info("Starting node...")

    prestart_checks(ctx, InstanceType.NODE, name, system_folders)

    context, namespace = select_context_and_namespace(
        context=context,
        namespace=namespace,
    )

    create_directory_if_not_exists(ctx.log_dir)
    create_directory_if_not_exists(ctx.data_dir)

    # Determine image-name. First we check if the image is set in the config file.
    # If so, we use it. Otherwise, we use the same version as the server.
    if custom_image := ctx.config.get("node", {}).get("image"):
        image = custom_image
    else:
        # if no custom image is specified, find the server version and use
        # the latest images from that minor version
        client = create_client(ctx)
        major_minor = None
        try:
            # try to get server version, skip if can't get a connection
            version = client.util.get_server_version(attempts_on_timeout=3)["version"]
            major_minor = ".".join(version.split(".")[:2])
            image = (
                f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_NODE_IMAGE_WO_TAG}:{major_minor}"
            )
        except Exception:
            warning("Could not determine server version. Using default node image")

        if major_minor and not __version__.startswith(major_minor):
            warning(
                "Version mismatch between CLI and server/node. CLI is running on "
                f"version {__version__}, while node and server are on version "
                f"{major_minor}. This might cause unexpected issues; changing to "
                f"{major_minor}.<latest> is recommended."
            )

    # fail safe, in case no custom image is specified and we can't get the
    # server version
    if not image:
        image = f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_NODE_IMAGE}"

    helm_install(
        release_name=ctx.helm_release_name,
        chart_name=ChartName.NODE,
        values_file=ctx.config_file,
        context=context,
        namespace=namespace,
        local_chart_dir=local_chart_dir,
        custom_values=[f"node.image={image}"],
    )

    # start port forward for the node proxy server
    start_port_forward(
        service_name=f"{ctx.helm_release_name}-node-service",
        service_port=ctx.config["node"].get("proxyPort", DEFAULT_PROXY_SERVER_PORT),
        port=ctx.config["node"].get("proxyPort", DEFAULT_PROXY_SERVER_PORT),
        context=context,
        namespace=namespace,
    )

    if attach:
        attach_logs(
            name,
            instance_type=InstanceType.NODE,
            infra_component=InfraComponentName.NODE,
            system_folders=system_folders,
            context=context,
            namespace=namespace,
            is_sandbox=ctx.is_sandbox,
        )



from __future__ import annotations

import re
import subprocess
import time
from os import PathLike
from pathlib import Path

from vantage6.common import error, info
from vantage6.common.context import AppContext
from vantage6.common.globals import (
    DEFAULT_CHART_REPO,
    LOCALHOST,
    InstanceType,
)

from vantage6.cli.common.utils import check_running
from vantage6.cli.globals import ChartName
from vantage6.cli.utils import check_config_name_allowed, validate_input_cmd_args


def prestart_checks(
    ctx: AppContext,
    instance_type: InstanceType,
    name: str,
    system_folders: bool,
) -> None:
    """
    Run pre-start checks for an instance.
    """

    check_config_name_allowed(name)

    if check_running(ctx.helm_release_name, instance_type, name, system_folders):
        error(f"Instance '{name}' is already running.")
        exit(1)


def helm_install(
    release_name: str,
    chart_name: ChartName,
    values_file: str | PathLike | None = None,
    context: str | None = None,
    namespace: str | None = None,
    local_chart_dir: str | None = None,
    custom_values: list[str] | None = None,
) -> None:
    """
    Manage the `helm install` command.

    Parameters
    ----------
    release_name : str
        The name of the Helm release.
    chart_name : str
        The name of the Helm chart.
    values_file : str, optional
        A single values file to use with the `-f` flag.
    context : str, optional
        The Kubernetes context to use.
    namespace : str, optional
        The Kubernetes namespace to use.
    local_chart_dir : str, optional
        The local directory containing the Helm charts.
    custom_values : list[str], optional
        Custom values to pass to the Helm chart, that override the values in the values
        file. Each item in the list is a string in the format "key=value".
    """
    # Input validation
    validate_input_cmd_args(release_name, "release name")
    validate_input_cmd_args(chart_name, "chart name")

    values_file = Path(values_file) if values_file else None
    if values_file and not values_file.is_file():
        error(f"Helm chart values file does not exist: {values_file}")
        return

    validate_input_cmd_args(context, "context name", allow_none=True)
    validate_input_cmd_args(namespace, "namespace name", allow_none=True)

    if local_chart_dir and local_chart_dir.rstrip("/").endswith(chart_name.value):
        local_chart_dir = str(Path(local_chart_dir).parent)

    # Create the command
    if local_chart_dir:
        command = [
            "helm",
            "install",
            release_name,
            f"{local_chart_dir}/{chart_name.value}",
        ]
    else:
        command = [
            "helm",
            "install",
            release_name,
            chart_name,
            "--repo",
            DEFAULT_CHART_REPO,
            # TODO v5+ remove this flag when we have a stable release
            "--devel",
        ]

    if values_file:
        command.extend(["-f", str(values_file)])

    if context:
        command.extend(["--kube-context", context])

    if namespace:
        command.extend(["--namespace", namespace])

    if custom_values:
        for custom_value in custom_values:
            command.extend(["--set", custom_value])

    try:
        subprocess.run(
            command,
            stdout=subprocess.DEVNULL,
            check=True,
        )
        info(
            f"Successfully installed release '{release_name}' using chart "
            f"'{chart_name.value}'."
        )
    except subprocess.CalledProcessError:
        error(f"Failed to install release '{release_name}'.")
        exit(1)
    except FileNotFoundError:
        error(
            "Helm command not found. Please ensure Helm is installed and available in "
            "the PATH."
        )
        exit(1)


def start_port_forward(
    service_name: str,
    service_port: int,
    port: int,
    ip: str = LOCALHOST,
    context: str | None = None,
    namespace: str | None = None,
) -> None:
    """
    Port forward a kubernetes service.

    Parameters
    ----------
    service_name : str
        The name of the Kubernetes service to port forward.
    service_port : int
        The port on the service to forward.
    port : int
        The port to listen on.
    ip : str
        The IP address to listen on. Defaults to localhost.
    context : str | None
        The Kubernetes context to use.
    namespace : str | None
        The Kubernetes namespace to use.
    """
    # Input validation
    validate_input_cmd_args(service_name, "service name")
    if not isinstance(service_port, int) or service_port <= 0:
        error(f"Invalid service port: {service_port}. Must be a positive integer.")
        return

    if not isinstance(port, int) or port <= 0:
        error(f"Invalid local port: {port}. Must be a positive integer.")
        return

    if ip and not re.match(
        r"^(localhost|[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})$", ip
    ):
        error(f"Invalid IP address: {ip}. Must be a valid IPv4 address or 'localhost'.")
        return

    validate_input_cmd_args(context, "context name", allow_none=True)
    validate_input_cmd_args(namespace, "namespace name", allow_none=True)

    # Check if the service is ready before starting port forwarding
    info(f"Waiting for service '{service_name}' to become ready...")
    start_time = time.time()
    timeout = 300  # seconds
    while time.time() - start_time < timeout:
        try:
            command = [
                "kubectl",
                "get",
                "endpoints",
                service_name,
                "-o",
                "jsonpath={.subsets[*].addresses[*].ip}",
            ]

            if context:
                command.extend(["--context", context])

            if namespace:
                command.extend(["--namespace", namespace])

            result = subprocess.check_output(command).decode().strip()

            if result:
                info(f"Service '{service_name}' is ready.")
                break
        except subprocess.CalledProcessError:
            pass  # ignore and retry

        time.sleep(2)
    else:
        error(
            f"Timeout: Service '{service_name}' has no ready endpoints after {timeout} "
            "seconds."
        )
        return

    # Create the port forwarding command
    if not ip:
        ip = LOCALHOST

    command = [
        "kubectl",
        "port-forward",
        "--address",
        ip,
        f"service/{service_name}",
        f"{port}:{service_port}",
    ]

    if context:
        command.extend(["--context", context])

    if namespace:
        command.extend(["--namespace", namespace])

    # Start the port forwarding process
    try:
        process = subprocess.Popen(
            command,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            start_new_session=True,  # Start in new session to detach from parent
        )

        # Give the process a moment to start and check if it's still running
        time.sleep(1)
        if process.poll() is not None:
            # Process has already terminated
            e = process.stderr.read().decode() if process.stderr else "Unknown error"
            error(f"Failed to start port forwarding: {e}")
            return

        info(
            f"Port forwarding started: {ip}:{port} -> {service_name}:{service_port} "
            f"(PID: {str(process.pid)})"
        )
        return
    except Exception as e:
        error(f"Failed to start port forwarding: {e}")
        return
